import csv
import itertools
from typing import Dict

from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta.commands.formats import Format
from spinta.utils.nestedstruct import flatten
from spinta.components import Context, Action, UrlParams, Model
from spinta import commands
from spinta.utils.response import aiter


class IterableFile:

    def __init__(self):
        self.writes = []

    def __iter__(self):
        yield from self.writes
        self.writes = []

    def write(self, data):
        self.writes.append(data)


class Csv(Format):
    content_type = 'text/csv'
    accept_types = {
        'text/csv',
    }
    params = {}

    def __call__(self, data):
        rows = flatten(data)
        peek = next(rows, None)

        if peek is None:
            return

        cols = list(peek.keys())
        rows = itertools.chain([peek], rows)

        stream = IterableFile()
        writer = csv.DictWriter(stream, fieldnames=cols)
        writer.writeheader()
        for row in rows:
            writer.writerow(row)
            yield from stream


@commands.render.register()
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Csv,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Dict[str, str] = None,
):
    headers = headers or {}
    headers['Content-Disposition'] = f'attachment; filename="{model.basename}.csv"'
    return StreamingResponse(
        aiter(fmt(data)),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )



from typing import List

from spinta import commands
from spinta import spyna
from spinta.core.ufuncs import asttoexpr
from spinta.datasets.components import Attribute
from spinta.datasets.helpers import load_resource_backend
from spinta.dimensions.lang.helpers import load_lang_data
from spinta.nodes import get_node, load_node
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.datasets.components import Dataset, Resource, Entity
from spinta.core.access import load_access_param
from spinta.types.namespace import load_namespace_from_name
from spinta.utils.data import take


@commands.load.register(Context, Dataset, dict, Manifest)
def load(
    context: Context,
    dataset: Dataset,
    data: dict,
    manifest: Manifest,
    *,
    source: Manifest = None,
):
    config = context.get('config')

    load_node(context, dataset, data, parent=manifest)
    load_access_param(dataset, data.get('access'), (manifest,))

    ns = load_namespace_from_name(context, manifest, dataset.name, drop=False)
    if ns.generated:
        ns.title = dataset.title
        ns.description = dataset.description

    dataset.lang = load_lang_data(context, dataset.lang)

    # Load resources
    dataset.resources = {}
    for name, params in (data.get('resources') or {}).items():
        resource = get_node(config, manifest, dataset.eid, data, parent=dataset, group='datasets', ctype='resource')
        resource.type = params.get('type')
        resource.name = name
        resource.dataset = dataset
        dataset.resources[name] = load(context, resource, params, manifest)

    return dataset


@commands.load.register(Context, Resource, dict, Manifest)
def load(context: Context, resource: Resource, data: dict, manifest: Manifest):
    load_node(context, resource, data, parent=resource.dataset)
    if resource.prepare:
        formula = resource.prepare
        if formula and isinstance(formula, str):
            # If formula is a string, then convert it to parse tree. Formula can
            # be given as string or parse tree.
            formula = spyna.parse(formula)
        resource.prepare = asttoexpr(formula)
    load_access_param(resource, data.get('access'), (resource.dataset,))
    resource.lang = load_lang_data(context, resource.lang)
    resource.backend = load_resource_backend(
        context,
        resource,
        # First backend is loaded as string and later becomes Backend.
        resource.backend,
    )

    # Models will be added on `link` command.
    resource.models = {}
    return resource


@commands.load.register(Context, Entity, dict, Manifest)
def load(context: Context, entity: Entity, data: dict, manifest: Manifest):
    # XXX: https://gitlab.com/atviriduomenys/spinta/-/issues/44
    pkeys: List[str] = entity.pkeys or []
    if pkeys:
        entity.pkeys = [entity.model.properties[k] for k in pkeys]
    else:
        entity.unknown_primary_key = True
        entity.pkeys = sorted(
            take(entity.model.properties).values(),
            key=lambda p: p.place,
        )

    if entity.prepare:
        entity.prepare = asttoexpr(entity.prepare)
    return entity


@commands.load.register(Context, Attribute, dict, Manifest)
def load(context: Context, attr: Attribute, data: dict, manifest: Manifest):
    if attr.prepare:
        attr.prepare = asttoexpr(attr.prepare)
    return attr



from typing import List, Dict, Any

import datetime

from authlib.oauth2 import rfc6749
from starlette.requests import Request

from spinta import commands
from spinta.components import Context, Config


class AccessLog:
    accessors: List[Dict[str, str]]
    method: str = None
    reason: str = None
    url: str = None
    buffer_size: int = 100

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return None

    def create_message(
        self,
        txn: str,
        method: str,
        reason: str,
        resources: List[Dict[str, Any]],
        fields: List[str],
    ):
        message = {
            'accessors': self.accessors,
            'http_method': method or self.method,
            'url': self.url,
            'reason': reason or self.reason,
            'timestamp': datetime.datetime.now(datetime.timezone.utc).isoformat(),
            'transaction_id': txn,
            'resources': resources,
            'fields': fields,
        }
        return message

    def log(
        self,
        *,
        txn: str = None,
        method: str = None,
        reason: str = None,
        resources: List[Dict[str, Any]],
        fields: List[str],
    ):
        raise NotImplementedError


@commands.load.register(Context, AccessLog, Config)
def load(context: Context, accesslog: AccessLog, config: Config):
    accesslog.buffer_size = config.rc.get('accesslog', 'buffer_size', required=True)


@commands.load.register(Context, AccessLog, rfc6749.TokenMixin)
def load(context: Context, accesslog: AccessLog, token: rfc6749.TokenMixin):  # noqa
    accesslog.accessors.extend([
        {
            'type': 'person',
            'id': token.get_sub(),
        },
        {
            'type': 'client',
            'id': token.get_aud(),
        },
    ])


@commands.load.register(Context, AccessLog, Request)
def load(context: Context, accesslog: AccessLog, request: Request):  # noqa
    accesslog.method = request.method
    accesslog.url = str(request.url)


def create_accesslog(
    context: Context,
    *,
    method: str = None,
    loaders: List[Any],
) -> AccessLog:
    config = context.get('config')
    accesslog = config.AccessLog()
    accesslog.method = method
    accesslog.accessors = []
    for loader in loaders:
        commands.load(context, accesslog, loader)
    return accesslog



import pathlib
from typing import Optional

from spinta import commands
from spinta.components import Context
from spinta.exceptions import ManifestFileDoesNotExist
from spinta.manifests.tabular.components import TabularManifest


@commands.configure.register(Context, TabularManifest)
def configure(context: Context, manifest: TabularManifest):
    rc = context.get('rc')
    path: Optional[str] = rc.get('manifests', manifest.name, 'path')
    if path:
        if (
            not path.startswith(('http://', 'https://')) and
            not pathlib.Path(path).exists()
        ):
            raise ManifestFileDoesNotExist(manifest, path=path)
    else:
        path = None
    manifest.path = path



from typing import List
from typing import Optional

from spinta import exceptions
from spinta.auth import authorized
from spinta.backends import Backend
from spinta.backends.components import BackendOrigin
from spinta.backends.helpers import load_backend
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import ShortExpr
from spinta.datasets.components import Resource
from spinta.dimensions.enum.helpers import get_prop_enum
from spinta.types.datatype import Ref
from spinta.ufuncs.changebase.helpers import change_base_model
from spinta.ufuncs.components import ForeignProperty
from spinta.ufuncs.helpers import merge_formulas
from spinta.utils.data import take
from spinta.utils.naming import to_code_name


def load_resource_backend(
    context: Context,
    resource: Resource,
    name: Optional[str],  # Backend name given in `ref` column.
) -> Backend:
    if name:
        # If backend name is given, just use an exiting backend.
        store = resource.dataset.manifest.store
        possible_backends = [
            # Backend defined in manifest.
            resource.dataset.manifest.backends,
            # Backend defined via configuration.
            store.backends,
        ]
        for backends in possible_backends:
            if name in backends:
                return backends[name]
        raise exceptions.BackendNotFound(resource, name=name)

    elif resource.type or resource.external:
        # If backend is defined inline on resource itself, try to load it.
        name = f'{resource.dataset.name}/{resource.name}'
        return load_backend(context, resource, name, BackendOrigin.resource, {
            'type': resource.type,
            'name': name,
            'dsn': resource.external,
        })

    else:
        # If backend was not given, try to check configuration with resource and
        # dataset names as backend names or fallback to manifest backend.
        store = resource.dataset.manifest.store
        possible_backends = [
            # Dataset and resource name combination.
            f'{resource.dataset.name}/{resource.name}',
            # Just dataset name.
            resource.dataset.name,
            # Just resource type.
            # XXX: I don't think that using resource type as a backend name is a
            #      good idea. Probably should be removed.
            resource.type,
        ]
        for name in filter(None, possible_backends):
            name = to_code_name(name)
            if name in store.backends:
                return store.backends[name]
        return resource.dataset.manifest.backend


def get_enum_filters(
    context: Context,
    model: Model,
    fpr: ForeignProperty = None,
) -> Optional[Expr]:
    args: List[Expr] = []
    for prop in take(['_id', all], model.properties).values():
        if (
            (enum := get_prop_enum(prop)) and
            authorized(context, prop, Action.GETALL)
        ):
            if not all(item.access >= prop.access for item in enum.values()):
                values = []
                for item in enum.values():
                    if item.access >= prop.access:
                        values.append(item.prepare)

                if fpr:
                    dtype = fpr.push(prop)
                else:
                    dtype = prop.dtype

                if len(values) == 1:
                    arg = values[0]
                elif len(values) > 1:
                    arg = values
                else:
                    arg = None

                if arg is not None:
                    expr = ShortExpr('eq', dtype.get_bind_expr(), arg)
                    args.append(expr)
    if len(args) == 1:
        return args[0]
    elif len(args) > 1:
        return ShortExpr('and', *args)
    else:
        return None


def get_ref_filters(
    context: Context,
    model: Model,
    fpr: ForeignProperty = None,
) -> Optional[Expr]:
    query: Optional[Expr] = None
    if fpr:
        query = merge_formulas(
            query,
            change_base_model(context, model, fpr),
        )
        query = merge_formulas(
            query,
            get_enum_filters(context, model, fpr),
        )

    for prop in take(['_id', all], model.properties).values():
        if (
            isinstance(prop.dtype, Ref) and
            authorized(context, prop, Action.GETALL)
        ):
            if fpr:
                fpr = fpr.push(prop)
            else:
                fpr = ForeignProperty(fpr, prop.dtype)

            query = merge_formulas(
                query,
                get_ref_filters(context, prop.dtype.model, fpr),
            )

    return query



import ujson as json

from typing import Optional

from starlette.requests import Request
from starlette.responses import JSONResponse, Response, StreamingResponse

from spinta.commands.formats import Format
from spinta.components import Context, Action, UrlParams, Node
from spinta import commands
from spinta.utils.response import aiter, peek_and_stream


class Json(Format):
    content_type = 'application/json'
    accept_types = {
        'application/json',
    }
    params = {}
    container_name = '_data'

    def __call__(self, data):
        yield f'{{"{self.container_name}":['
        for i, row in enumerate(data):
            sep = ',' if i > 0 else ''
            yield sep + json.dumps(self.data(row), ensure_ascii=False)
        yield ']}'

    def data(self, data: dict) -> dict:
        return data


@commands.render.register()  # noqa
def render(
    context: Context,
    request: Request,
    node: Node,
    fmt: Json,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(
        context,
        request,
        fmt,
        action,
        params,
        data,
        status_code,
        headers,
    )


def _render(
    context: Context,
    request: Request,
    fmt: Json,
    action: Action,
    params: UrlParams,
    data,
    status_code: int,
    headers: dict,
):
    if action in (Action.GETALL, Action.SEARCH, Action.CHANGES):
        # In python dict is also an iterable, but here we want a true iterable,
        # a list or a generator of dicts.
        assert not isinstance(data, dict), data
        return StreamingResponse(
            aiter(peek_and_stream(fmt(data))),
            status_code=status_code,
            media_type=fmt.content_type,
            headers=headers,
        )
    elif action == Action.DELETE:
        return Response(None, status_code=status_code, headers=headers)
    else:
        return JSONResponse(
            fmt.data(data),
            status_code=status_code,
            headers=headers
        )



from typing import Any
from typing import AsyncIterator, Union, List, Dict, Optional, Iterable, Generator

import base64
import datetime
import decimal
import pathlib
import uuid

from spinta import spyna
from spinta import commands
from spinta import exceptions
from spinta.auth import authorized
from spinta.commands import load_operator_value, prepare, gen_object_id, is_object_id
from spinta.components import Context, Namespace, Model, Property, Action, Node, DataItem
from spinta.exceptions import ConflictingValue, NoItemRevision
from spinta.types.datatype import DataType, DateTime, Date, Object, Array, String, File, PrimaryKey, Binary, Ref, JSON, Number
from spinta.utils.itertools import chunks
from spinta.utils.schema import NotAvailable, NA
from spinta.utils.data import take
from spinta.backends.components import Backend

SelectTree = Optional[Dict[str, dict]]


@commands.prepare_for_write.register(Context, Model, Backend, dict)
def prepare_for_write(
    context: Context,
    model: Model,
    backend: Backend,
    patch: Dict[str, Any],
    *,
    action: Action,
) -> dict:
    # prepares model's data for storing in a backend
    backend = model.backend
    result = {}
    for name, value in patch.items():
        if not name.startswith('_'):
            prop = model.properties[name]
            value = commands.prepare_for_write(context, prop.dtype, backend, value)
        result[name] = value
    return result


@commands.prepare_for_write.register(Context, Property, Backend, object)
def prepare_for_write(
    context: Context,
    prop: Property,
    backend: Backend,
    value: Any,
    *,
    action: Action,
) -> Any:
    if prop.name in value:
        value[prop.name] = commands.prepare_for_write(
            context,
            prop.dtype,
            prop.dtype.backend,
            value[prop.name],
        )
    return value


@commands.prepare_for_write.register(Context, DataType, Backend, object)
def prepare_for_write(
    context: Context,
    dtype: DataType,
    backend: Backend,
    value: Any,
) -> Any:
    # prepares value for data store
    # for simple types - loaded native values should work
    # otherwise - override for this command if necessary
    return value


@commands.prepare_for_write.register(Context, Array, Backend, list)
def prepare_for_write(
    context: Context,
    dtype: Array,
    backend: Backend,
    value: List[Any],
) -> List[Any]:
    # prepare array and it's items for datastore
    return [
        commands.prepare_for_write(context, dtype.items.dtype, backend, v)
        for v in value
    ]


@commands.prepare_for_write.register(Context, Object, Backend, dict)
def prepare_for_write(
    context: Context,
    dtype: Object,
    backend: Backend,
    value: Dict[str, Any],
) -> Dict[str, Any]:
    # prepare objects and it's properties for datastore
    prepped = {}
    for k, v in value.items():
        prop = dtype.properties[k]
        prepped[k] = commands.prepare_for_write(context, prop.dtype, backend, v)
    return prepped


@prepare.register(Context, Backend, Property)
def prepare(context: Context, backend: Backend, prop: Property):
    return prepare(context, backend, prop.dtype)


@commands.simple_data_check.register(Context, DataItem, Model, Backend)
def simple_data_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    simple_model_properties_check(context, data, model, backend)


@commands.simple_data_check.register(Context, DataItem, Property, Backend)
def simple_data_check(
    context: Context,
    data: DataItem,
    prop: Property,
    backend: Backend,
) -> None:
    simple_data_check(
        context,
        data,
        prop.dtype,
        prop,
        backend,
        data.given[prop.name],
    )


def simple_model_properties_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    for name, prop in model.properties.items():
        simple_data_check(
            context,
            data,
            prop.dtype,
            prop,
            prop.dtype.backend,
            data.given.get(name, NA),
        )


@commands.simple_data_check.register(Context, DataItem, DataType, Property, Backend, object)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: object,
) -> None:
    check_type_value(dtype, value)

    # Action.DELETE is ignore for qvarn compatibility reasons.
    # XXX: make `spinta` check for revision on Action.DELETE,
    #      implementers can override this command as they please.
    updating = data.action in (Action.UPDATE, Action.PATCH)
    if updating and '_revision' not in data.given:
        raise NoItemRevision(prop)


@commands.simple_data_check.register(Context, DataItem, Object, Property, Backend, dict)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: object,
) -> None:
    for prop in dtype.properties.values():
        v = value.get(prop.name, NA)
        simple_data_check(context, data, prop.dtype, prop, prop.dtype.backend, v)


@commands.simple_data_check.register(Context, DataItem, Array, Property, Backend, list)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: list,
) -> None:
    dtype = dtype.items.dtype
    for v in value:
        simple_data_check(context, data, dtype, prop, dtype.backend, v)


@commands.complex_data_check.register(Context, DataItem, Model, Backend)
def complex_data_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    # XXX: Maybe Action.DELETE should also provide revision.
    updating = data.action in (Action.UPDATE, Action.PATCH)
    if updating and '_revision' not in data.given:
        raise NoItemRevision(model)

    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    model.properties[k],
                    given=data.given[k],
                    expected=data.saved[k],
                )

    complex_model_properties_check(context, model, backend, data)


def complex_model_properties_check(
    context: Context,
    model: Model,
    backend: Backend,
    data: DataItem,
) -> None:
    for name, prop in model.properties.items():
        # For datasets, property type is optional.
        # XXX: But I think, it should be mandatory.
        if prop.dtype is not None:
            given = data.given.get(name, NA)

            if prop.dtype.unique and given is not NA:
                commands.check_unique_constraint(
                    context,
                    data,
                    prop.dtype,
                    prop,
                    data.backend,
                    given,
                )
            complex_data_check(
                context,
                data,
                prop.dtype,
                prop,
                data.backend,
                given,
            )


@complex_data_check.register(Context, DataItem, DataType, Property, Backend, object)
def complex_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: object,
):
    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    dtype,
                    given=data.given[k],
                    expected=data.saved[k],
                )


def check_type_value(dtype: DataType, value: object):
    if dtype.required and (value is None or value is NA):
        # FIXME: Raise a UserError
        raise Exception(f"{dtype.prop.name!r} is required for {dtype.prop.model.name!r}.")


@gen_object_id.register(Context, Backend, Node)
def gen_object_id(context: Context, backend: Backend, model: Node):
    return str(uuid.uuid4())


@is_object_id.register(Context, str)
def is_object_id(context: Context, value: str):
    # Collect all available backend/model combinations.
    # TODO: this probably should be cached at load time to increase performance
    #       of object id lookup.
    # XXX: other option would be to implement two pass URL parsing, to get
    #      information about model, and then call is_object_id directly with
    #      backend and model. That would be nice.
    store = context.get('store')
    candidates = set()
    manifest = store.manifest
    for model in manifest.objects.get('model', {}).values():
        candidates.add((model.backend.__class__, model.__class__))

    # Currently all models use UUID, but dataset models use id generated from
    # other properties that form an unique id.
    for Backend_, Model_ in candidates:
        backend = Backend_()
        model = Model_()
        model.name = ''
        if is_object_id(context, backend, model, value):
            return True


@is_object_id.register(Context, Backend, Model, str)
def is_object_id(context: Context, backend: Backend, model: Model, value: str):
    try:
        return uuid.UUID(value).version == 4
    except ValueError:
        return False


@is_object_id.register(Context, Backend, Model, object)
def is_object_id(context: Context, backend: Backend, model: Model, value: object):
    # returns false if object id is non string.
    # strings are handled by other command
    return False


@is_object_id.register(Context, type(None), Namespace, object)
def is_object_id(context: Context, backend: type(None), model: Namespace, value: object):
    # Namespaces do not have object id.
    return False


@commands.prepare_data_for_response.register(Context, Action, Model, Backend, dict)
def prepare_data_for_response(
    context: Context,
    action: Action,
    model: Model,
    backend: Backend,
    value: dict,
    *,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            backend,
            model,
            prop.dtype,
            val,
            select=sel,
        )
        for prop, val, sel in _select_model_props(
            model,
            prop_names,
            value,
            select,
            ['_type', '_id', '_revision'],
        )
    }


@commands.prepare_data_for_response.register(Context, Action, Object, Backend, dict)
def prepare_data_for_response(
    context: Context,
    action: Action,
    dtype: Object,
    backend: Backend,
    value: dict,
    *,
    select: List[str] = None,
) -> dict:
    select = flat_select_to_nested(select)
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            backend,
            prop.model,
            prop.dtype,
            value,
            select=select,
        )
        for prop, value, select in _select_prop_props(
            dtype.prop,
            value,
            select,
            ['_type', '_revision'],
        )
    }


@commands.prepare_data_for_response.register(Context, Action, File, Backend, dict)
def prepare_data_for_response(
    context: Context,
    action: Action,
    dtype: File,
    backend: Backend,
    value: dict,
    *,
    select: List[str] = None,
    property_: Property = None,
) -> dict:
    prop = dtype.prop
    reserved = ['_type', '_revision']

    if prop.name in value and value[prop.name]:
        check_unknown_props(prop, select, set(reserved) | set(value[prop.name]))
    else:
        check_unknown_props(prop, select, set(reserved))

    if prop.name in value and value[prop.name]:
        data = commands.prepare_dtype_for_response(
            context,
            backend,
            prop.model,
            prop.dtype,
            value[prop.name],
            select=select,
        )
    else:
        data = {}

    for key, val, sel in _select_props(prop, reserved, None, value, select):
        data[key] = val

    return data


def _apply_always_show_id(
    context: Context,
    action: Action,
    select: Optional[List[str]],
) -> Optional[List[str]]:
    if action in (Action.GETALL, Action.SEARCH):
        config = context.get('config')
        if config.always_show_id:
            if select is None:
                return ['_id']
            elif '_id' not in select:
                return ['_id'] + select
    return select


def get_select_tree(
    context: Context,
    action: Action,
    select: Optional[List[dict]],
) -> SelectTree:
    select = _apply_always_show_id(context, action, select)
    if select is None and action in (Action.GETALL, Action.SEARCH):
        # If select is not given, select everything.
        select = {'*': {}}
    return flat_select_to_nested(select)


def get_select_prop_names(
    context: Context,
    model: Model,
    action: Action,
    select: SelectTree,
):
    reserved = {'_type', '_id', '_revision'}
    check_unknown_props(model, select, reserved | set(take(model.properties)))

    if select is None or '*' in select:
        return [
            p.name
            for p in model.properties.values() if (
                not p.name.startswith('_') and
                not p.hidden and
                authorized(context, p, action)
            )
        ]
    else:
        return [k for k in select if not k.startswith('_')]


def _select_model_props(
    model: Model,
    prop_names: List[str],
    value: dict,
    select: SelectTree,
    reserved: List[str],
):
    yield from _select_props(
        model,
        reserved,
        model.properties,
        value,
        select,
    )
    yield from _select_props(
        model,
        prop_names,
        model.properties,
        value,
        select,
        reserved=False,
    )


def _select_prop_props(
    prop: Property,
    value: dict,
    select: SelectTree,
    reserved: List[str],
):
    if prop.name in value:
        check_unknown_props(prop, select, set(reserved) | set(value[prop.name]))
    else:
        check_unknown_props(prop, select, set(reserved))
    yield from _select_props(
        prop,
        reserved,
        prop.model.properties,
        value,
        select,
    )
    if prop.name in value:
        yield from _select_props(
            prop,
            value[prop.name].keys(),
            prop.dtype.properties,
            value[prop.name],
            select,
        )


def _select_props(
    node: Union[Namespace, Model, Property],
    keys: Iterable[str],
    props: Optional[Dict[str, Property]],
    value: dict,
    select: SelectTree,
    *,
    reserved: bool = True,
):
    for k in keys:

        if select is None and k not in value:
            # Omit all keys if they are not present in value, this is a common
            # case in PATCH requests.
            continue

        if reserved is False and k.startswith('_'):
            continue

        if select is None:
            sel = None
        elif '*' in select:
            sel = select['*']
        elif k in select:
            sel = select[k]
        else:
            continue

        if sel is not None and sel == {}:
            sel = {'*': {}}

        if props is None:
            prop = k
        else:
            if k not in props:
                # FIXME: We should check select list at the very beginning of
                #        request, not when returning results.
                raise exceptions.FieldNotInResource(node, property=k)
            prop = props[k]
            if prop.hidden:
                continue

        if k in value:
            val = value[k]
        else:
            val = None

        yield prop, val, sel


# FIXME: We should check select list at the very beginning of
#        request, not when returning results.
def check_unknown_props(
    node: Union[Model, Property, DataType],
    select: Optional[Iterable[str]],
    known: Iterable[str],
):
    unknown_properties = set(select or []) - set(known) - {'*'}
    if unknown_properties:
        raise exceptions.MultipleErrors(
            exceptions.FieldNotInResource(node, property=prop)
            for prop in sorted(unknown_properties)
        )


def flat_select_to_nested(select: Optional[List[str]]) -> SelectTree:
    """
    >>> flat_select_to_nested(None)

    >>> flat_select_to_nested(['foo.bar'])
    {'foo': {'bar': {}}}

    """
    if select is None:
        return None

    res = {}
    for v in select:
        if isinstance(v, dict):
            v = spyna.unparse(v)
        names = v.split('.')
        vref = res
        for name in names:
            if name not in vref:
                vref[name] = {}
            vref = vref[name]

    return res


@commands.prepare_dtype_for_response.register(Context, Backend, Model, DataType, object)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: DataType,
    value: object,
    *,
    select: dict = None,
):
    assert isinstance(value, (str, int, float, bool, type(None))), (
        f"prepare_dtype_for_response must return only primitive, json "
        f"serializable types, {type(value)} is not a primitive data type, "
        f"model={model!r}, dtype={dtype!r}"
    )
    return value


@commands.prepare_dtype_for_response.register(Context, Backend, Model, DataType, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: DataType,
    value: NotAvailable,
    *,
    select: dict = None,
):
    return dtype.default


@commands.prepare_dtype_for_response.register(Context, Backend, Model, File, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: File,
    value: NotAvailable,
    *,
    select: dict = None,
):
    return {
        '_id': None,
        '_content_type': None,
    }


@commands.prepare_dtype_for_response.register(Context, Backend, Model, File, dict)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: File,
    value: dict,
    *,
    select: dict = None,
):
    data = {
        key: val
        for key, val, sel in _select_props(
            dtype.prop,
            ['_id', '_content_type'],
            None,
            value,
            select,
        )
    }

    if isinstance(data.get('_id'), pathlib.Path):
        # _id on FileSystem backend is a Path.
        data['_id'] = str(data['_id'])

    # File content is returned only if explicitly requested.
    if select and '_content' in select:
        if '_content' in value:
            content = value['_content']
        else:
            prop = dtype.prop
            content = commands.getfile(context, prop, dtype, dtype.backend, data=value)

        if content is None:
            data['_content'] = None
        else:
            data['_content'] = base64.b64encode(content).decode()

    return data


@commands.prepare_dtype_for_response.register(Context, Backend, Model, DateTime, datetime.datetime)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: DateTime,
    value: datetime.datetime,
    *,
    select: dict = None,
):
    return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Date, datetime.date)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Date,
    value: datetime.date,
    *,
    select: dict = None,
):
    if isinstance(value, datetime.datetime):
        return value.date().isoformat()
    else:
        return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Binary, bytes)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Binary,
    value: bytes,
    *,
    select: dict = None,
):
    return base64.b64encode(value).decode('ascii')


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Ref, dict)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Ref,
    value: dict,
    *,
    select: dict = None,
):
    if select is None and value == {'_id': None}:
        return None
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            backend,
            prop.model,
            prop.dtype,
            val,
            select=sel,
        )
        for prop, val, sel in _select_ref_props(
            dtype,
            value,
            select,
            ['_id'],
        )
    }


def _select_ref_props(
    dtype: Ref,
    value: dict,
    select: SelectTree,
    reserved: List[str],
):
    yield from _select_props(
        dtype.prop,
        reserved,
        dtype.model.properties,
        value,
        select,
    )
    yield from _select_props(
        dtype.model,
        value.keys(),
        dtype.model.properties,
        value,
        select,
    )


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Ref, str)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Ref,
    value: str,
    *,
    select: dict = None,
):
    # FIXME: Backend should never return references as strings! References
    #        should always be dicts.
    return {'_id': value}


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Object, dict)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Object,
    value: dict,
    *,
    select: dict = None,
):
    check_unknown_props(dtype, select, set(take(dtype.properties)))

    if select is None:
        keys = value.keys()
    elif '*' in select:
        keys = take(dtype.properties).keys()
    else:
        keys = [k for k in select if not k.startswith('_')]

    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            backend,
            model,
            prop.dtype,
            val,
            select=sel,
        )
        for prop, val, sel in _select_props(
            dtype.prop,
            keys,
            dtype.properties,
            value,
            select,
        )
    }


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Array, list)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Array,
    value: list,
    *,
    select: dict = None,
) -> list:
    return _prepare_array_for_response(
        context,
        backend,
        model,
        dtype,
        value,
        select,
    )


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Array, tuple)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Array,
    value: tuple,
    *,
    select: dict = None,
) -> list:
    return _prepare_array_for_response(
        context,
        backend,
        model,
        dtype,
        value,
        select,
    )


def _prepare_array_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Array,
    value: Iterable,
    select: dict = None,
) -> list:
    if dtype.items:
        return [
            commands.prepare_dtype_for_response(
                context,
                backend,
                model,
                dtype.items.dtype,
                v,
                select=select,
            )
            for v in value
        ]
    else:
        return [v for v in value]


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Array, type(None))
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Array,
    value: type(None),
    *,
    select: dict = None,
):
    return []


@commands.prepare_dtype_for_response.register(Context, Backend, Model, JSON, object)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: JSON,
    value: object,
    *,
    select: dict = None,
):
    return value


@commands.prepare_dtype_for_response.register(Context, Backend, Model, JSON, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: JSON,
    value: NotAvailable,
    *,
    select: dict = None,
):
    return None


@commands.prepare_dtype_for_response.register(Context, Backend, Model, Number, decimal.Decimal)
def prepare_dtype_for_response(
    context: Context,
    backend: Backend,
    model: Model,
    dtype: Number,
    value: decimal.Decimal,
    *,
    select: dict = None,
):
    return float(value)


@commands.unload_backend.register(Context, Backend)
def unload_backend(context: Context, backend: Backend):
    pass


@load_operator_value.register(Context, Backend, DataType, object)
def load_operator_value(
    context: Context,
    backend: Backend,
    dtype: DataType,
    value: object,
    *,
    query_params: dict,
):
    operator = query_params['name']
    # XXX: That is probably not a very reliable way of getting original operator
    #      name. Maybe at least this should be documented some how.
    operator_name = query_params.get('origin', operator)
    if operator in ('startswith', 'contains') and not isinstance(dtype, (String, PrimaryKey)):
        raise exceptions.InvalidOperandValue(dtype, operator=operator_name)
    if operator in ('gt', 'ge', 'lt', 'le') and isinstance(dtype, String):
        raise exceptions.InvalidOperandValue(dtype, operator=operator_name)


@commands.make_json_serializable.register(Model, dict)
def make_json_serializable(model: Model, value: dict) -> dict:
    return commands.make_json_serializable[Object, dict](model, value)


@commands.make_json_serializable.register(DataType, object)
def make_json_serializable(dtype: DataType, value: object):
    return value


@commands.make_json_serializable.register(DateTime, datetime.datetime)
def make_json_serializable(dtype: DateTime, value: datetime.datetime):
    return value.isoformat()


@commands.make_json_serializable.register(Date, datetime.date)
def make_json_serializable(dtype: Date, value: datetime.date):
    return value.isoformat()


@commands.make_json_serializable.register(Date, datetime.datetime)
def make_json_serializable(dtype: Date, value: datetime.datetime):
    return value.date().isoformat()


def _get_json_serializable_props(dtype: Object, value: dict):
    for k, v in value.items():
        if k in dtype.properties:
            yield dtype.properties[k], v
        else:
            Exception(f"Unknown {k} key in {dtype!r} object.")


@commands.make_json_serializable.register(Object, dict)
def make_json_serializable(dtype: Object, value: dict):
    return {
        prop.name: commands.make_json_serializable(prop.dtype, v)
        for prop, v in _get_json_serializable_props(dtype, value)
    }


@commands.make_json_serializable.register(Array, list)
def make_json_serializable(dtype: Array, value: list):
    return [make_json_serializable(dtype.items.dtype, v) for v in value]


@commands.update.register(Context, Property, DataType, Backend)
def update(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


@commands.patch.register(Context, Property, DataType, Backend)
def patch(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


@commands.delete.register(Context, Property, DataType, Backend)
def delete(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    dstream = _prepare_property_for_delete(prop, dstream)
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


async def _prepare_property_for_delete(
    prop: Property,
    dstream: AsyncIterator[DataItem],
) -> AsyncIterator[DataItem]:
    async for data in dstream:
        if data.patch:
            data.patch = {
                **{k: v for k, v in data.patch.items() if k.startswith('_')},
                prop.name: None,
            }
        yield data


@commands.cast_backend_to_python.register(Context, Model, Backend, dict)
def cast_backend_to_python(
    context: Context,
    model: Model,
    backend: Backend,
    data: dict,
) -> dict:
    return {
        k: commands.cast_backend_to_python(
            context,
            model.properties[k].dtype,
            backend,
            v,
        ) if k in model.properties else v
        for k, v in data.items()
    }


@commands.cast_backend_to_python.register(Context, Property, Backend, object)
def cast_backend_to_python(
    context: Context,
    prop: Property,
    backend: Backend,
    data: object,
) -> dict:
    return commands.cast_backend_to_python(context, prop.dtype, backend, data)


@commands.cast_backend_to_python.register(Context, DataType, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: DataType,
    backend: Backend,
    data: object,
) -> dict:
    return data


@commands.cast_backend_to_python.register(Context, Object, Backend, dict)
def cast_backend_to_python(
    context: Context,
    dtype: Object,
    backend: Backend,
    data: dict,
) -> dict:
    return {
        k: commands.cast_backend_to_python(
            context,
            dtype.properties[k].dtype,
            backend,
            v,
        ) if k in dtype.properties else v
        for k, v in data.items()
    }


@commands.cast_backend_to_python.register(Context, Array, Backend, list)
def cast_backend_to_python(
    context: Context,
    dtype: Array,
    backend: Backend,
    data: list,
) -> dict:
    return [
        commands.cast_backend_to_python(
            context,
            dtype.items.dtype,
            backend,
            v,
        )
        for v in data
    ]


@commands.cast_backend_to_python.register(Context, Binary, Backend, str)
def cast_backend_to_python(
    context: Context,
    dtype: Binary,
    backend: Backend,
    data: str,
) -> dict:
    return base64.b64decode(data)


def log_getall(
    context: Context,
    rows: Generator[dict, None, None],
    select: List[dict] = None,
    hidden: List[str] = None,
):
    accesslog = context.get('accesslog')
    for chunk in chunks(rows, accesslog.buffer_size):
        resources = [
            {
                '_id': data.get('_id'),
                '_type': data.get('_type'),
                '_revision': data.get('_revision')
            } for data in chunk
        ]
        fields = _get_selected_fields(select)
        accesslog.log(resources=resources, fields=fields)
        yield from chunk


def log_getone(
    context: Context,
    result: dict,
    select: List[str] = None,
    hidden: List[str] = None,
):
    fields = _get_selected_fields(select)
    # XXX: select queries do not return metadata
    resource = {
        '_id': result.get('_id'),  # XXX: subresource queries do not return id
        '_type': result.get('_type'),
        '_revision': result.get('_revision'),
    }
    accesslog = context.get('accesslog')
    accesslog.log(resources=[resource], fields=fields)


def _get_selected_fields(select: Optional[List[dict]]) -> List[str]:
    if select:
        return sorted([_get_selected_field(x) for x in select])
    else:
        return []


def _get_selected_field(ast: dict):
    if ast['name'] == 'bind':
        return ast['args'][0]
    if ast['name'] == 'getattr':
        return '.'.join([
            _get_selected_field(ast['args'][0]),
            _get_selected_field(ast['args'][1]),
        ])



from typing import Dict

from spinta import commands
from spinta import exceptions
from spinta.backends import Backend
from spinta.backends.components import BackendOrigin
from spinta.components import Component
from spinta.components import Context


def load_backend(
    context: Context,
    component: Component,
    name: str,
    origin: BackendOrigin,
    data: Dict[str, str]
) -> Backend:
    config = context.get('config')
    type_ = data.get('type')
    if not type_:
        raise exceptions.RequiredConfigParam(
            component,
            name=f'backends.{name}.type',
        )
    if type_ not in config.components['backends']:
        raise exceptions.BackendNotFound(component, name=type_)
    Backend_ = config.components['backends'][type_]
    backend: Backend = Backend_()
    backend.type = type_
    backend.name = name
    backend.origin = origin
    backend.config = data
    commands.load(context, backend, data)
    return backend



from typing import Tuple

from spinta import commands
from spinta.cli.helpers.store import load_store
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.helpers import load_ascii_tabular_manifest
from spinta.manifests.tabular.helpers import normalizes_columns
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.manifests.tabular.helpers import striptable
from spinta.testing.context import create_test_context


def compare_manifest(manifest: Manifest, expected: str) -> Tuple[str, str]:
    expected = striptable(expected)
    if expected:
        header = expected.splitlines()[0]
        cols = normalizes_columns(header.split('|'))
        sizes = {c: len(x) - 2 for c, x in zip(cols, f' {header} '.split('|'))}
        actual = render_tabular_manifest(manifest, cols, sizes=sizes)
    else:
        actual = render_tabular_manifest(manifest)
    return actual, expected


def configure_manifest(rc: RawConfig, manifest: str) -> Context:
    rc = rc.fork({
        'keymaps.test': {
            'type': 'sqlalchemy',
            'dsn': 'sqlite://',
        },
        'backends': {
            'memory': {
                'type': 'memory',
            },
        },
        'manifests.test': {
            'type': 'tabular',
            'backend': 'memory',
            'keymap': 'test',
            'mode': 'external',
            'path': None,
        },
        'manifest': 'test',
    })
    context = create_test_context(rc)
    store = load_store(context)
    load_ascii_tabular_manifest(context, store.manifest, manifest, strip=True)
    commands.check(context, store.manifest)
    commands.prepare(context, store.manifest)
    context.loaded = True
    return context



from starlette.requests import Request

from spinta import commands
from spinta.components import Context, Node, Action, UrlParams


@commands.getall.register(Context, Request, Node)
async def getall(
    context: Context,
    request: Request,
    node: Node,
    *,
    action: Action,
    params: UrlParams,
):
    return await commands.getall(context, request, node, node.backend, action=action, params=params)


@commands.getone.register(Context, Request, Node)
async def getone(
    context: Context,
    request: Request,
    node: Node,
    *,
    action: Action,
    params: UrlParams,
):
    if params.prop and params.propref:
        return await commands.getone(context, request, params.prop, params.model.backend, action=action, params=params)
    elif params.prop:
        return await commands.getone(context, request, params.prop, params.prop.dtype.backend or params.model.backend, action=action, params=params)
    else:
        return await commands.getone(context, request, params.model, params.model.backend, action=action, params=params)


@commands.changes.register()
async def changes(
    context: Context,
    request: Request,
    node: Node,
    *,
    action: Action,
    params: UrlParams,
):
    return await commands.changes(context, request, node, node.backend, action=action, params=params)



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import contextlib

from spinta import commands
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Node
from spinta.components import Context
from spinta.auth import AdminToken
from spinta.utils.imports import importstr
from spinta.core.context import create_context
from spinta.core.config import RawConfig


def create_test_context(
    rc: RawConfig,
    *args,
    name: str = 'pytest',
    **kwargs,
) -> TestContext:
    rc = rc.fork()
    Context_ = rc.get('components', 'core', 'context', cast=importstr)
    Context_ = type('ContextForTests', (ContextForTests, Context_), {})
    context = Context_(name)
    return create_context(name, rc, context)


class ContextForTests:
    loaded: bool = False

    def __init__(
        self: TestContext,
        name: str,
        parent: TestContext = None,
    ):
        super().__init__(name, parent)
        self.loaded = parent.loaded if parent else False

    @contextlib.contextmanager
    def transaction(self: TestContext, *, write=False):
        if self.has('transaction'):
            yield self
        else:
            with self:
                store = self.get('store')
                self.set('auth.token', AdminToken())
                backend = store.manifest.backend
                self.attach('transaction', backend.transaction, write=write)
                yield self

    def wipe(self: TestContext, model: Union[str, Node]):
        if isinstance(model, str):
            store = self.get('store')
            model = store.manifest.models[model]
        with self.transaction() as context:
            commands.wipe(context, model, model.backend)

    def wipe_all(self: TestContext):
        store = self.get('store')
        self.wipe(store.manifest.objects['ns'][''])

    def load(
        self: TestContext,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> TestContext:
        # We pass context to tests unloaded, by doing this, we give test
        # functions opportunity to call `context.load` manually and provide
        # `overrides` for config, this way each test can configure context in
        # anyway they want.
        #
        # If test function does not explicitly call `context.load`, then it will
        # be called implicitly on `app.request` and on some context methods,
        # that run database queries.
        if self.loaded:
            raise Exception("test context is already loaded")

        rc = self.get('rc')

        if overrides:
            rc.add('test', {
                'environments': {
                    'test': overrides,
                }
            })

        store = prepare_manifest(self)
        commands.bootstrap(self, store.manifest)

        self.loaded = True

        return self


TestContext = Union[Context, ContextForTests]



import ujson as json

from typing import Optional

from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta.commands.formats import Format
from spinta.components import Context, Action, UrlParams, Model, Node
from spinta import commands


class JsonLines(Format):
    content_type = 'application/x-json-stream'
    accept_types = {
        'application/x-json-stream',
    }
    params = {}

    def __call__(self, data):
        for row in data:
            yield json.dumps(row, ensure_ascii=False) + '\n'


@commands.render.register()  # noqa
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: JsonLines,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(fmt, data, status_code, headers)


@commands.render.register()  # noqa
def render(  # noqa
    context: Context,
    request: Request,
    node: Node,
    fmt: JsonLines,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(fmt, data, status_code, headers)


def _render(fmt: JsonLines, data, status_code: int, headers: dict):
    stream = _stream(data)
    return StreamingResponse(
        stream,
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers
    )


async def _stream(data):
    for row in data:
        yield json.dumps(row, ensure_ascii=False) + '\n'



import logging

from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.tabular.components import TabularManifest
from spinta.manifests.tabular.helpers import read_tabular_manifest

log = logging.getLogger(__name__)


@commands.load.register(Context, TabularManifest)
def load(
    context: Context,
    manifest: TabularManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
):
    assert freezed, (
        "TabularManifest does not have unfreezed version of manifest."
    )

    target = into or manifest
    if '_schema' not in target.models:
        store = context.get('store')
        commands.load(context, store.internal, into=target)

    if manifest.path is None:
        # There are cases, when we don't want to load any manifest.
        # For example, we want to start with an empty manifest, and then
        # generate it from external resources.
        return

    if into:
        log.info(
            'Loading freezed manifest %r into %r from %s.',
            manifest.name,
            into.name,
            manifest.path,
        )
        schemas = read_tabular_manifest(manifest.path)
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        log.info(
            'Loading freezed manifest %r from %s.',
            manifest.name,
            manifest.path,
        )
        schemas = read_tabular_manifest(manifest.path)
        load_manifest_nodes(context, manifest, schemas)



from typing import TextIO, List, Dict, Any, Union

import json
import sys
import pathlib

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.components import Context, Config, Store


class FileAccessLog(AccessLog):
    file: Union[TextIO, pathlib.Path]
    close: bool

    def __exit__(self, *exc):
        if self.close:
            self.file.close()

    def log(
        self,
        *,
        txn: str = None,
        method: str = None,
        reason: str = None,
        resources: List[Dict[str, Any]],
        fields: List[str],
    ):
        if self.file is None:
            return
        message = self.create_message(txn, method, reason, resources, fields)
        message = json.dumps(message)
        print(message, file=self.file)


@commands.load.register(Context, FileAccessLog, Config)
def load(context: Context, accesslog: FileAccessLog, config: Config):
    commands.load[Context, AccessLog, Config](context, accesslog, config)

    file = config.rc.get('accesslog', 'file')
    close = False

    if file == 'stdout':
        file = sys.stdout
    elif file == 'stderr':
        file = sys.sterr
    elif file == '/dev/null':
        file = None
    else:
        close = True
        file = pathlib.Path(file)

        # Try to open file to check if it is writable.
        with file.open('a'):
            pass

    accesslog.file = file
    accesslog.close = close
    return accesslog


@commands.load.register(Context, FileAccessLog, Store)
def load(context: Context, accesslog: FileAccessLog, store: Store):  # noqa
    file = store.accesslog.file
    if isinstance(file, pathlib.Path):
        file = file.open('a')
    accesslog.file = file
    accesslog.close = store.accesslog.close



import cgi
import tempfile
from typing import AsyncGenerator
from typing import AsyncIterator
from typing import Dict, Any

import boto3
import botocore

from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta import commands
from spinta.backends import Backend, simple_data_check, log_getone
from spinta.backends.components import BackendFeatures
from spinta.commands import getall
from spinta.commands.write import prepare_patch, simple_response, validate_data, log_write
from spinta.components import Action, Context, DataItem, Property, UrlParams
from spinta.exceptions import ItemDoesNotExist
from spinta.manifests.components import Manifest
from spinta.renderer import render
from spinta.types.datatype import File
from spinta.utils.aiotools import aiter
from spinta.utils.data import take


class S3(Backend):

    bucket_name: str = None
    region: str = None

    features = {
        BackendFeatures.WRITE,
    }


@commands.load.register(Context, S3, dict)
def load(context: Context, backend: S3, config: Dict[str, Any]):
    backend.bucket_name = config['bucket']
    backend.region = config['region']
    backend.credentials = {
        'aws_access_key_id': config.get('access_key_id'),
        'aws_secret_access_key': config.get('secret_access_key'),
    }


@commands.wait.register(Context, S3)
def wait(context: Context, backend: S3, *, fail: bool = False):
    try:
        # create s3 client and check if we can list our buckets (a way to
        # make requests to s3 to make sure that aws credentials are ok)
        client = get_aws_session(backend).client('s3')
        client.list_buckets()
        return True
    except botocore.exceptions.ClientError:
        return False


@commands.prepare.register()
def prepare(context: Context, backend: S3, manifest: Manifest):
    client = get_aws_session(backend).client('s3')
    try:
        client.get_bucket_location(Bucket=backend.bucket_name)
    except client.exceptions.NoSuchBucket:
        client.create_bucket(
            ACL='private',
            Bucket=backend.bucket_name,
            CreateBucketConfiguration={'LocationConstraint': backend.region},
        )


@commands.bootstrap.register()
def bootstrap(context: Context, backend: S3):
    pass


@commands.migrate.register(Context, S3)
def migrate(context: Context, backend: S3):
    pass


@commands.push.register()
async def push(
    context: Context,
    request: Request,
    dtype: File,
    backend: S3,
    *,
    action: Action,
    params: UrlParams,
):
    prop = dtype.prop
    commands.authorize(context, action, prop)
    data = DataItem(
        prop.model,
        prop,
        propref=False,
        backend=backend,
        action=action
    )
    data.given = {
        prop.name: {
            '_content_type': request.headers.get('content-type'),
            '_id': None,
        }
    }

    if 'revision' in request.headers:
        data.given['_revision'] = request.headers.get('revision')
    if 'content-disposition' in request.headers:
        data.given[prop.name]['_id'] = cgi.parse_header(request.headers['content-disposition'])[1]['filename']

    if not data.given[prop.name]['_id']:
        # XXX: Probably here should be a new UUID.
        data.given[prop.name]['_id'] = params.pk

    simple_data_check(context, data, data.prop, data.model.backend)

    data.saved = getone(context, prop, dtype, prop.model.backend, id_=params.pk)

    dstream = aiter([data])
    dstream = validate_data(context, dstream)
    dstream = prepare_patch(context, dstream)
    dstream = log_write(context, dstream)
    filename = data.given[prop.name]['_id']
    if action == Action.UPDATE:
        if 'content-length' not in request.headers:
            raise HTTPException(status_code=411)
        dstream = upload_file_to_s3(backend, filename, dstream, request.stream())
        dstream = commands.update(context, prop, dtype, prop.model.backend, dstream=dstream)
        dstream = commands.create_changelog_entry(
            context, prop.model, prop.model.backend, dstream=dstream,
        )
    elif action == Action.DELETE:
        dstream = commands.delete(context, prop, dtype, prop.model.backend, dstream=dstream)
        dstream = commands.create_changelog_entry(
            context, prop.model, prop.model.backend, dstream=dstream,
        )
    else:
        raise Exception(f"Unknown action {action!r}.")

    status_code, response = await simple_response(context, dstream)
    return render(context, request, prop, params, response, action=action, status_code=status_code)


async def upload_file_to_s3(
    backend: S3,
    filename: str,
    dstream: AsyncIterator[DataItem],
    fstream: AsyncIterator[bytes]
) -> AsyncGenerator[DataItem, None]:
    aws_session = get_aws_session(backend)
    s3_bucket = aws_session.resource('s3').Bucket(backend.bucket_name)

    async for d in dstream:
        with tempfile.TemporaryFile() as f:
            async for chunk in fstream:
                f.write(chunk)
            f.seek(0)
            s3_bucket.upload_fileobj(f, filename)
        yield d


@commands.getone.register()
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: File,
    backend: S3,
    *,
    action: str,
    params: UrlParams,
):
    commands.authorize(context, action, prop)
    data = getone(context, prop, prop.dtype, prop.model.backend, id_=params.pk)
    log_getone(context, data)
    value = data[prop.name]
    filename = value['_id']
    if filename is None:
        raise ItemDoesNotExist(prop, id=params.pk)

    client = get_aws_session(backend).client('s3')
    obj = client.get_object(Bucket=backend.bucket_name, Key=filename)
    data_gen = obj['Body'].iter_chunks()

    return StreamingResponse(
        aiter(data_gen),
        media_type=value.get('_content_type'),
        headers={
            'Revision': data['_revision'],
            'Content-Disposition': f'attachment; filename="{filename}"',
        },
    )


@commands.wipe.register(Context, File, S3)
def wipe(context: Context, dtype: File, backend: S3):
    aws = get_aws_session(backend)
    objs = aws.client('s3').list_objects(Bucket=backend.bucket_name)

    if 'Contents' in objs:
        obj_keys = {'Objects': []}
        fnames = [obj['Key'] for obj in objs['Contents']]
        rows = getall(context, dtype.prop.model, dtype.prop.model.backend)
        for r in rows:
            fn = take(dtype.prop.name + '._id', r)
            if fn and fn in fnames:
                obj_keys['Objects'].append({'Key': fn})
        if obj_keys['Objects']:
            bucket = aws.resource('s3').Bucket(backend.bucket_name)
            bucket.delete_objects(Delete=obj_keys)


def get_aws_session(backend: Backend, **kwargs):
    return boto3.Session(region_name=backend.region,
                         **backend.credentials,
                         **kwargs)



import configparser
import pathlib
import urllib.parse
from typing import List
from typing import NamedTuple

import requests

from spinta.exceptions import RemoteClientCredentialsNotFound
from spinta.exceptions import RemoteClientCredentialsNotGiven
from spinta.exceptions import RemoteClientScopesNotGiven


class _ParsedUrl(NamedTuple):
    section: str
    client: str
    secret: str
    server: str


def _parse_url(url: str) -> _ParsedUrl:
    if url.startswith(('https://', 'https://')):
        url_ = url
    else:
        url_ = 'https://' + url

    purl = urllib.parse.urlparse(url_)
    section = purl.hostname
    if purl.port:
        section += f':{purl.port}'
    if purl.username:
        section = f'{purl.username}@{purl.hostname}'
    client = purl.username
    secret = purl.password
    server = f'{purl.scheme}://{purl.hostname}'
    return _ParsedUrl(section, client, secret, server)


def add_client_credentials(
    credsfile: pathlib.Path,
    server: str,
    *,
    client: str = None,
    secret: str = None,
    scopes: List[str] = None,
    section: str = None,
) -> None:
    creds = configparser.ConfigParser()
    if credsfile.exists():
        creds.read(credsfile)

    purl = _parse_url(server)
    scopes = scopes or []
    scopes = '\n' + '\n'.join(f'  {s}' for s in scopes)

    section = section or purl.section

    creds[section] = {
        'server': purl.server,
        'client': client or purl.client or '',
        'secret': secret or purl.secret or '',
        'scopes': scopes,

    }

    with credsfile.open('w') as f:
        creds.write(f)


def get_access_token(url: str, credsfile: pathlib.Path) -> str:
    purl: _ParsedUrl = _parse_url(url)

    if not credsfile.exists():
        raise RemoteClientCredentialsNotFound(
            url=url,
            section=purl.section,
            credentials=credsfile,
        )

    creds = configparser.ConfigParser()
    creds.read(credsfile)

    if not creds.has_section(purl.section):
        raise RemoteClientCredentialsNotFound(
            url=url,
            section=purl.section,
            credentials=credsfile,
        )

    server = creds.get(purl.section, 'server', fallback=purl.server)
    client = creds.get(purl.section, 'client', fallback=purl.client)
    secret = purl.secret or creds.get(purl.section, 'secret', fallback=None)

    if not client or not secret:
        raise RemoteClientCredentialsNotGiven(
            url=url,
            section=purl.section,
            credentials=credsfile,
        )

    scopes = creds.get(purl.section, 'scopes', fallback=None)
    if not scopes:
        raise RemoteClientScopesNotGiven(
            url=url,
            section=purl.section,
            credentials=credsfile,
        )

    auth = (client, secret)
    resp = requests.post(f'{server}/auth/token', auth=auth, data={
        'grant_type': 'client_credentials',
        'scope': scopes,
    })
    resp.raise_for_status()
    return resp.json()['access_token']



from typing import List

import types
import itertools


def flatten(value, sep='.'):
    value, lists = _flatten(value, sep)

    if value is None:
        for k, vals in lists:
            for v in vals:
                yield from flatten(v, sep)

    elif lists:
        keys, lists = zip(*lists)
        for vals in itertools.product(*lists):
            val = {
                sep.join(k): v
                for k, v in zip(keys, vals)
            }
            val.update(value)
            yield from flatten(val, sep)

    else:
        yield value


def _flatten(value, sep, key=()):
    if isinstance(value, dict):
        data = {}
        lists = []
        for k, v in value.items():
            v, more = _flatten(v, sep, key + (k,))
            data.update(v or {})
            lists += more
        return data, lists

    elif isinstance(value, (
        list,
        types.GeneratorType,
        itertools.groupby,
        # _grouper is type of `g` in `k, g = next(itertools.groupby([1, 1]))`
        itertools._grouper,
        itertools.chain,
    )):
        if value:
            return None, [(key, value)]
        else:
            return None, []

    else:
        return {sep.join(key): value}, []


def build_select_tree(select: List[str]):
    tree = {}
    for name in select:
        split = name, None
        while len(split) == 2:
            name, node = split
            if name not in tree:
                tree[name] = set()
            if node:
                tree[name].add(node)
            split = name.rsplit('.', 1)
    return tree



import itertools
import operator
import subprocess
import sys

from typing import Optional

from starlette.requests import Request
from starlette.responses import StreamingResponse

from texttable import Texttable

from spinta.commands.formats import Format
from spinta.utils.nestedstruct import flatten
from spinta.components import Context, Action, UrlParams, Model
from spinta import commands
from spinta.utils.response import aiter


class Ascii(Format):
    content_type = 'text/plain'
    accept_types = {
        'text/plain',
    }
    params = {
        'width': {'type': 'integer'},
        'colwidth': {'type': 'integer'},
    }

    def __call__(self, data, width=None, colwidth=42):
        data = iter(data)
        peek = next(data, None)

        if peek is None:
            return

        data = itertools.chain([peek], data)

        if '_type' in peek:
            groups = itertools.groupby(data, operator.itemgetter('_type'))
        else:
            groups = [(None, data)]

        for name, group in groups:
            if name:
                yield f'\n\nTable: {name}\n'

            rows = flatten(group)
            peek = next(rows, None)
            rows = itertools.chain([peek], rows)

            if name:
                cols = [k for k in peek.keys() if k != '_type']
            else:
                cols = list(peek.keys())

            if colwidth:
                width = len(cols) * colwidth

            buffer = [cols]
            tnum = 1

            for row in rows:
                buffer.append([row.get(c) for c in cols])
                if len(buffer) > 100:
                    yield from _draw(buffer, name, tnum, width)
                    buffer = [cols]
                    tnum += 1

            if buffer:
                yield from _draw(buffer, name, tnum, width)


@commands.render.register()  # noqa
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Ascii,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(fmt, params, data, status_code, headers)


def _render(fmt: Ascii, params: UrlParams, data, status_code, headers):
    # Format params ar given in RQL query `?format(width(1),colwidth(1))`.
    width = params.formatparams.get('width')
    colwidth = params.formatparams.get('colwidth')

    if width is None and colwidth is None and sys.stdin.isatty():
        _, width = map(int, subprocess.run(['stty', 'size'], capture_output=True).stdout.split())
    elif width is None and colwidth is None:
        colwidth = 42

    return StreamingResponse(
        aiter(fmt(data, width, colwidth)),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )


def _draw(buffer, name, tnum, width):
    if tnum > 1:
        if name:
            yield f"\n\nTable {name} #{tnum}:\n"
        else:
            yield f"\n\nTable #{tnum}:\n"

    table = Texttable(width)
    table.set_deco(Texttable.HEADER)
    table.add_rows(buffer)
    yield table.draw()



from spinta.components import Action


class Format:
    content_type: str
    accept_types: set
    params: dict
    # TODO: Should be false, by default.
    streamable: bool = True

    async def __call__(self, rows, action: Action, **params):
        raise NotImplementedError



import csv
import pathlib
from io import StringIO
from typing import List
from typing import Literal
from typing import Union
from typing import overload

from spinta import commands
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.tabular.components import ManifestColumn
from spinta.manifests.tabular.constants import DATASET
from spinta.manifests.tabular.helpers import read_ascii_tabular_manifest
from spinta.manifests.tabular.helpers import read_ascii_tabular_rows
from spinta.manifests.tabular.helpers import read_tabular_manifest
from spinta.manifests.tabular.helpers import torow
from spinta.manifests.tabular.helpers import write_tabular_manifest
from spinta.testing.context import create_test_context


def create_tabular_manifest(
    path: pathlib.Path,
    manifest: str,
):
    path = pathlib.Path(path)
    rows = read_ascii_tabular_rows(manifest, strip=True)
    cols: List[ManifestColumn] = next(rows, [])
    if cols:
        rows = (torow(DATASET, dict(zip(cols, row))) for row in rows)
        write_tabular_manifest(str(path), rows)


@overload
def load_tabular_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    *,
    return_context: Literal[True],
) -> Context:
    ...


@overload
def load_tabular_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    *,
    return_context: Literal[False],
) -> Manifest:
    ...


def load_tabular_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    *,
    return_context: bool = False,
) -> Union[Manifest, Context]:
    if isinstance(manifest, str) and '|' in manifest:
        path = ''
    else:
        path = str(manifest)
        manifest = None

    rc = rc.fork({
        'manifest': 'default',
        'manifests': {
            'default': {
                'type': 'tabular',
                'path': path,
                'keymap': 'default',
            },
        },
    })

    context = create_test_context(rc)

    config = context.get('config')
    commands.load(context, config)

    store = context.get('store')
    commands.load(context, store)

    if path:
        schemas = read_tabular_manifest(path)
    else:
        schemas = read_ascii_tabular_manifest(manifest, strip=True)
    load_manifest_nodes(context, store.manifest, schemas)

    commands.link(context, store.manifest)

    if return_context:
        return context
    else:
        return store.manifest


def convert_ascii_manifest_to_csv(manifest: str) -> str:
    file = StringIO()
    writer = csv.writer(file)
    rows = read_ascii_tabular_rows(
        manifest,
        strip=True,
        check_column_names=False,
    )
    for row in rows:
        writer.writerow(row)
    return file.getvalue()






from typing import Optional

import datetime
import uuid
import hashlib
import decimal

import msgpack
import sqlalchemy as sa

from spinta import commands
from spinta.components import Context
from spinta.datasets.keymaps.components import KeyMap


class SqlAlchemyKeyMap(KeyMap):
    dsn: str = None

    def __init__(self, dsn: str = None):
        self.dsn = dsn
        self.engine = None
        self.metadata = None
        self.conn = None

    def __enter__(self):
        assert self.dsn is not None
        assert self.conn is None
        self.conn = self.engine.connect()
        return self

    def __exit__(self, *exc):
        self.conn.close()
        self.conn = None

    def get_table(self, name):
        if name not in self.metadata.tables:
            table = sa.Table(
                name, self.metadata,
                sa.Column('key', sa.Text, primary_key=True),
                sa.Column('hash', sa.Text, unique=True),
                sa.Column('value', sa.LargeBinary),
            )
            table.create(checkfirst=True)
        return self.metadata.tables[name]

    def encode(self, name: str, value: object) -> Optional[str]:
        # Make value msgpack serializable.
        if isinstance(value, (list, tuple)):
            value = [_encode_value(k) for k in value if k is not None]
            if len(value) == 0:
                return None
        else:
            if value is None:
                return None
            value = _encode_value(value)

        # Get value hash.
        value = msgpack.dumps(value, strict_types=True)
        hash = hashlib.sha1(value).hexdigest()

        # Try to find key by value hash.
        table = self.get_table(name)
        query = sa.select([table.c.key]).where(table.c.hash == hash)
        key = self.conn.execute(query).scalar()
        if key is None:

            # Create new key.
            key = str(uuid.uuid4())
            self.conn.execute(table.insert(), {
                'key': key,
                'hash': hash,
                'value': value,
            })

        return key

    def decode(self, name: str, key: str) -> object:
        table = self.get_table(name)
        query = sa.select([table.c.value]).where(table.c.key == key)
        value = self.conn.execute(query).scalar()
        value = msgpack.loads(value, raw=False)
        return value


def _encode_value(value):
    if isinstance(value, (datetime.datetime, datetime.date)):
        return value.isoformat()
    if isinstance(value, decimal.Decimal):
        return float(value)
    if isinstance(value, tuple):
        return list(value)
    return value


@commands.configure.register(Context, SqlAlchemyKeyMap)
def configure(context: Context, keymap: SqlAlchemyKeyMap):
    rc = context.get('rc')
    keymap.dsn = rc.get('keymaps', keymap.name, 'dsn', required=True)


@commands.prepare.register(Context, SqlAlchemyKeyMap)
def prepare(context: Context, keymap: SqlAlchemyKeyMap):
    keymap.engine = sa.create_engine(keymap.dsn)
    keymap.metadata = sa.MetaData(keymap.engine)



from typing import Any
from typing import Dict
from typing import NamedTuple
from typing import Optional, Union, List

import datetime
from typing import Tuple

import pkg_resources as pres

from starlette.requests import Request
from starlette.templating import Jinja2Templates
from starlette.exceptions import HTTPException

from spinta.components import Config
from spinta.manifests.components import Manifest
from spinta.types.datatype import DataType, Ref
from spinta.commands.formats import Format
from spinta.components import Context, Action, UrlParams, Node, Property
from spinta.utils.url import build_url_path
from spinta import commands
from spinta.components import Namespace, Model
from spinta.utils.nestedstruct import flatten
from spinta.auth import authorized


class Html(Format):
    content_type = 'text/html'
    accept_types = {
        'text/html',
    }
    params = {}
    streamable = False


def _render_check(request: Request, data: Dict[str, Any] = None):
    if data:
        if 'errors' in data:
            result = {
                'message': "Yra klaidų",
                'errors': [
                    err['message']
                    for err in data['errors']
                ],
            }
        else:
            result = {
                'message': "Klaidų nerasta.",
                'errors': None,
            }
    else:
        result = data

    templates = Jinja2Templates(directory=pres.resource_filename('spinta', 'templates'))
    return templates.TemplateResponse('form.html', {
        'request': request,
        'title': "Duomenų struktūros aprašo tikrinimas",
        'description': (
            "Ši priemonė leidžia patikrinti ar "
            "<a href=\"https://atviriduomenys.readthedocs.io/dsa/index.html\">"
            "duomenų struktūros apraše</a> nėra klaidų."
        ),
        'name': 'check',
        'fields': [
            {
                'label': "Duomenų struktūros aprašas",
                'help': "Pateikite duomenų struktūros aprašo failą.",
                'input': '<input type="file" name="manifest" accept=".csv">'
            },
        ],
        'submit': "Tikrinti",
        'result': result,
    })



@commands.render.register()
def render(
    context: Context,
    request: Request,
    ns: Namespace,
    fmt: Html,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    if action == Action.CHECK:
        return _render_check(request, data)
    else:
        return _render_model(context, request, ns, action, params, data, headers)


@commands.render.register()  # noqa
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Html,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render_model(context, request, model, action, params, data, headers)


def _render_model(
    context: Context,
    request: Request,
    model: Union[Model, Namespace],
    action: Action,
    params: UrlParams,
    data,
    http_headers,
):
    header = []
    row = []

    if model.type == 'model:ns' or params.ns:
        data = get_ns_data(data)
        header = next(data)
        data = list(data)
    elif action == Action.CHANGES:
        data = get_changes(context, data, model, params)
        header = next(data)
        data = list(reversed(list(data)))
    elif action == Action.GETONE:
        row = list(get_row(context, data, model))
        data = []
    elif action in (Action.GETALL, Action.SEARCH):
        data = get_data(context, data, model, params, action)
        header = next(data)
        data = list(data)

    templates = Jinja2Templates(directory=pres.resource_filename('spinta', 'templates'))
    return templates.TemplateResponse(
        'base.html',
        {
            **get_template_context(context, model, params),
            'request': request,
            'header': header,
            'data': data,
            'row': row,
            'formats': get_output_formats(params),
            'limit_enforced': params.limit_enforced,
            'params': params,
        },
        headers=http_headers
    )


def get_output_formats(params: UrlParams):
    return [
        # XXX I don't like that, there should be a better way to build links
        #     from UrlParams instance.
        ('CSV', '/' + build_url_path(params.changed_parsetree({'format': ['csv']}))),
        ('JSON', '/' + build_url_path(params.changed_parsetree({'format': ['json']}))),
        ('JSONL', '/' + build_url_path(params.changed_parsetree({'format': ['jsonl']}))),
        ('ASCII', '/' + build_url_path(params.changed_parsetree({'format': ['ascii']}))),
    ]


def get_template_context(context: Context, model, params: UrlParams):
    config: Config = context.get('config')
    return {
        'location': get_current_location(config, model, params),
    }


CurrentLocation = List[Tuple[
    str,            # Link title
    Optional[str],  # Link URL
]]


class PathInfo(NamedTuple):
    path: str = ''
    name: str = ''
    link: str = ''
    title: str = ''


def _split_path(
    manifest: Manifest,
    base: str,
    orig_path: str,
) -> List[PathInfo]:
    parts = orig_path.split('/') if orig_path else []
    result: List[PathInfo] = []
    last = len(parts)
    base = [base] if base else []
    for i, part in enumerate(parts, 1):
        path = '/'.join(base + parts[:i])
        if i == last and path in manifest.models:
            title = manifest.models[path].title
        elif path in manifest.namespaces:
            title = manifest.namespaces[path].title
        else:
            title = ''
        title = title or part
        result.append(PathInfo(
            path=path,
            name=part,
            link=f'/{path}',
            title=title,
        ))
    return result


def get_current_location(
    config: Config,
    model: Model,
    params: UrlParams,
) -> CurrentLocation:
    # Remove config root
    path = params.path
    if config.root:
        if path.startswith(config.root):
            path = path[len(config.root) + 1:]
        elif config.root.startswith(path):
            path = ''

    parts = _split_path(model.manifest, config.root, path)
    if len(parts) > 0:
        parts, last = parts[:-1], parts[-1]
    else:
        parts, last = [], None

    loc: CurrentLocation = [('🏠', '/')]
    loc += [(p.title, p.link) for p in parts]

    if model.type == 'model:ns':
        if last is not None:
            loc += [(last.title, None)]

    else:
        pk = params.pk
        changes = params.changes

        if last is not None:
            if pk or changes:
                loc += [(last.title, get_model_link(model))]
            else:
                loc += [(last.title, None)]

        if pk:
            pks = params.pk[:8]  # short version of primary key
            if changes:
                loc += [(pks, get_model_link(model, pk=pk))]
            else:
                loc += [(pks, None)]

        if changes:
            loc += [('Changes', None)]
        else:
            loc += [('Changes', get_model_link(model, pk=pk, changes=[]))]

    return loc


def get_changes(context: Context, rows, model, params: UrlParams):
    props = [
        p for p in model.properties.values() if (
            not p.name.startswith('_')
        )
    ]

    yield (
        ['_id', '_revision', '_txn', '_created', '_op', '_rid'] +
        [prop.name for prop in props if prop.name != 'revision']
    )

    # XXX: With large changes sets this will consume a lot of memmory.
    current = {}
    for data in rows:
        id_ = data['_rid']
        if id_ not in current:
            current[id_] = {}
        current[id_].update({
            k: v for k, v in data.items() if not k.startswith('_')
        })
        row = [
            {'color': None, 'value': data['_id'], 'link': None},
            {'color': None, 'value': data['_revision'], 'link': None},
            {'color': None, 'value': data['_txn'], 'link': None},
            {'color': None, 'value': data['_created'], 'link': None},
            {'color': None, 'value': data['_op'], 'link': None},
            get_cell(context, model.properties['_id'], data, '_rid', shorten=True),
        ]
        for prop in props:
            if prop.name in data:
                color = 'change'
            elif prop.name not in current:
                color = 'null'
            else:
                color = None
            cell = get_cell(
                context,
                prop,
                current[id_],
                prop.name,
                shorten=True,
                color=color,
            )
            row.append(cell)
        yield row


def get_row(context: Context, row, model: Node):
    if row is None:
        raise HTTPException(status_code=404)
    include = {'_type', '_id', '_revision'}
    for prop in model.properties.values():
        if (
            not prop.hidden and
            # TODO: object and array are not supported yet
            prop.dtype.name not in ('object', 'array') and
            (prop.name in include or not prop.name.startswith('_'))
        ):
            yield prop.name, get_cell(context, prop, row, prop.name)


def get_cell(
    context: Context,
    prop: Property,
    row: Dict[str, Any],
    name: Any,
    shorten=False,
    color=None,
):
    COLORS = {
        'change': '#B2E2AD',
        'null': '#f5f5f5',
    }

    link = None
    model = None
    if prop.dtype.name == 'ref':
        value = row.get(f'{name}._id')
    else:
        value = row.get(name)

    if prop.name == '_id' and value:
        model = prop.model
    elif prop.dtype.name == 'ref' and prop.dtype.model and value:
        model = prop.dtype.model

    if model:
        link = '/' + build_url_path(get_model_link_params(model, pk=value))

    if prop.dtype.name in ('ref', 'pk') and shorten and isinstance(value, str):
        value = value[:8]

    if isinstance(value, datetime.datetime):
        value = value.isoformat()

    max_column_length = 200
    if shorten and isinstance(value, str) and len(value) > max_column_length:
        value = value[:max_column_length] + '...'

    if value is None:
        value = ''
        if color is None:
            color = 'null'

    return {
        'value': value,
        'link': link,
        'color': COLORS[color] if color else None,
    }


def get_ns_data(rows):
    yield ['title', 'description']
    for row in rows:
        if row['title']:
            title = row['title']
        else:
            parts = row['_id'].split('/')
            if row['_type'] == 'ns':
                title = parts[-2]
            else:
                title = parts[-1]

        if row['_type'] == 'ns':
            icon = '📁'
            suffix = '/'
        else:
            icon = '📄'
            suffix = ''

        yield [
            {
                'value': f'{icon} {title}{suffix}',
                'link': '/' + row['_id'],
                'color': None,
            },
            {
                'value': row['description'],
                'link': None,
                'color': None,
            },
        ]


def get_data(
    context: Context,
    rows,
    model: Model,
    params: UrlParams,
    action: Action,
):
    # XXX: For things like aggregations, a dynamic model should be created with
    #      all the properties coming from aggregates.
    if params.count:
        prop = Property()
        prop.dtype = DataType()
        prop.dtype.name = 'string'
        prop.name = 'count()'
        prop.ref = None
        prop.model = model
        props = [prop]
        header = ['count()']
    else:
        if params.select:
            header = [_expr_to_name(x) for x in params.select]
            props = _get_props_from_select(context, model, header)
        else:
            include = {'_id'}
            props = [
                p for p in model.properties.values() if (
                    authorized(context, p, action) and
                    not p.hidden and
                    # TODO: object and array are not supported yet
                    p.dtype.name not in ('object', 'array') and
                    (p.name in include or not p.name.startswith('_'))
                )
            ]
            header = [p.name for p in props]

    yield header

    for data in flatten(rows):
        row = []
        for name, prop in zip(header, props):
            row.append(get_cell(context, prop, data, name, shorten=True))
        yield row


def _get_props_from_select(
    context: Context,
    model: Model,
    select: List[str],
) -> List[Property]:
    props = []
    for node in select:
        name = _expr_to_name(node)
        if name in model.flatprops:
            prop = model.flatprops[name]
        else:
            parts = name.split('.')
            prop = _find_linked_prop(context, model, parts[0], parts[1:])
        props.append(prop)
    return props


def _expr_to_name(node) -> str:
    if not isinstance(node, dict):
        return node

    name = node['name']

    if name == 'bind':
        return node['args'][0]

    if name == 'getattr':
        obj, key = node['args']
        return _expr_to_name(obj) + '.' + _expr_to_name(key)

    raise Exception(f"Unknown node {node!r}.")


def _find_linked_prop(
    context: Context,
    model: Model,
    name: str,
    parts: List[str],
) -> Union[Property, None]:
    # TODO: Add support for nested properties, now only references are
    #       supported.
    prop = model.properties.get(name)
    if parts and prop and isinstance(prop.dtype, Ref):
        model = prop.dtype.model
        return _find_linked_prop(context, model, parts[0], parts[1:])
    else:
        return prop


def get_model_link_params(model: Node, *, pk: Optional[str] = None, **extra):
    ptree = [
        {
            'name': 'path',
            'args': (
                model.name.split('/') +
                ([pk] if pk is not None else [])
            ),
        }
    ]

    for k, v in extra.items():
        ptree.append({
            'name': k,
            'args': v,
        })

    return ptree


def get_model_link(*args, **kwargs):
    return '/' + build_url_path(get_model_link_params(*args, **kwargs))



from typing import Optional

from typer import Argument
from typer import Context as TyperContext
from typer import echo

from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.tabular.helpers import render_tabular_manifest


def show(
    ctx: TyperContext,
    manifest: Optional[str] = Argument(None),
):
    """Show manifest as ascii table"""

    context: Context = ctx.obj
    context = context.fork('inspect')

    if manifest is not None:
        config = {
            'keymaps.show': {
                'type': 'sqlalchemy',
                'dsn': 'sqlite://',
            },
            'backends.null': {
                'type': 'memory',
            },
            'manifests.show': {
                'type': 'tabular',
                'backend': 'null',
                'keymap': 'show',
                'mode': 'internal',
                'path': manifest,
            },
            'manifest': 'show',
        }

        # Add given manifest file to configuration
        rc: RawConfig = context.get('rc')
        context.set('rc', rc.fork(config))

    store = prepare_manifest(context, verbose=False)
    manifest = store.manifest
    echo(render_tabular_manifest(manifest))




from starlette.requests import Request

from spinta.backends import get_select_prop_names
from spinta.backends import get_select_tree
from spinta.compat import urlparams_to_expr
from spinta import commands
from spinta.renderer import render
from spinta.components import Context, Action, UrlParams
from spinta.components import Model
from spinta.backends import log_getall
from spinta.datasets.components import ExternalBackend


@commands.getall.register(Context, Request, Model, ExternalBackend)
async def getall(
    context: Context,
    request: Request,
    model: Model,
    backend: ExternalBackend,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)
    expr = urlparams_to_expr(params, add_count=False)
    rows = commands.getall(context, model, backend, query=expr)
    hidden_props = [prop.name for prop in model.properties.values() if prop.hidden]
    rows = log_getall(context, rows, select=params.select, hidden=hidden_props)
    if not params.count:
        select_tree = get_select_tree(context, action, params.select)
        prop_names = get_select_prop_names(context, model, action, select_tree)
        rows = (
            commands.prepare_data_for_response(
                context,
                action,
                model,
                backend,
                row,
                select=select_tree,
                prop_names=prop_names,
            )
            for row in rows
        )
    return render(context, request, model, params, rows, action=action)



from types import AsyncGeneratorType

import sqlalchemy as sa

from starlette.requests import Request

from spinta import commands
from spinta.renderer import render
from spinta.utils.json import fix_data_for_json
from spinta.components import Context, Action, UrlParams, Model, Property
from spinta.backends.postgresql.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.sqlalchemy import utcnow


@commands.create_changelog_entry.register(Context, (Model, Property), PostgreSQL)
async def create_changelog_entry(
    context: Context,
    node: (Model, Property),
    backend: PostgreSQL,
    *,
    dstream: AsyncGeneratorType,
) -> None:
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(node, TableType.CHANGELOG)
    async for data in dstream:
        qry = table.insert().values(
            _txn=transaction.id,
            datetime=utcnow(),
            action=Action.INSERT.value,
        )
        connection.execute(qry, [{
            '_rid': data.saved['_id'] if data.saved else data.patch['_id'],
            '_revision': data.patch['_revision'] if data.patch else data.saved['_revision'],
            '_txn': transaction.id,
            'datetime': utcnow(),
            'action': data.action.value,
            'data': fix_data_for_json({
                k: v for k, v in data.patch.items() if not k.startswith('_')
            }),
        }])
        yield data


@commands.changes.register()
async def changes(
    context: Context,
    request: Request,
    model: Model,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)
    data = changes(context, model, backend, id_=params.pk, limit=params.limit, offset=params.offset)
    data = (
        {
            **row,
            '_created': row['_created'].isoformat(),
        }
        for row in data
    )
    return render(context, request, model, params, data, action=action)


@changes.register()
def changes(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    id_: str = None,
    limit: int = 100,
    offset: int = -10,
):
    connection = context.get('transaction').connection
    table = backend.get_table(model, TableType.CHANGELOG)

    qry = sa.select([table]).order_by(table.c._id)
    qry = _changes_id(table, qry, id_)
    qry = _changes_offset(table, qry, offset)
    qry = _changes_limit(qry, limit)

    result = connection.execute(qry)
    for row in result:
        yield {
            '_id': row[table.c._id],
            '_revision': row[table.c._revision],
            '_txn': row[table.c._txn],
            '_rid': row[table.c._rid],
            '_created': row[table.c.datetime],
            '_op': row[table.c.action],
            **dict(row[table.c.data]),
        }


def _changes_id(table, qry, id_):
    if id_:
        return qry.where(table.c._rid == id_)
    else:
        return qry


def _changes_offset(table, qry, offset):
    if offset:
        if offset > 0:
            offset = offset
        else:
            offset = (
                qry.with_only_columns([
                    sa.func.max(table.c.change) - abs(offset),
                ]).
                order_by(None).alias()
            )
        return qry.where(table.c.change > offset)
    else:
        return qry


def _changes_limit(qry, limit):
    if limit:
        return qry.limit(limit)
    else:
        return qry



from starlette.requests import Request
from starlette.responses import FileResponse

from spinta import commands
from spinta.components import Context, Property, UrlParams
from spinta.backends import log_getone
from spinta.backends.fs.components import FileSystem
from spinta.types.datatype import DataType, File
from spinta.exceptions import ItemDoesNotExist


@commands.getone.register(Context, Request, Property, File, FileSystem)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: File,
    backend: FileSystem,
    *,
    action: str,
    params: UrlParams,
):
    commands.authorize(context, action, prop)
    data = getone(context, prop, prop.dtype, prop.model.backend, id_=params.pk)
    log_getone(context, data)
    value = data[prop.name]
    filename = value['_id']
    if filename is None:
        raise ItemDoesNotExist(prop, id=params.pk)
    return FileResponse(
        str(dtype.backend.path / filename),
        media_type=value.get('_content_type'),
        headers={
            'Revision': data['_revision'],
            'Content-Disposition': (
                f'attachment; filename="{filename}"'
                if filename else
                'attachment'
            )
        },
    )


@commands.getone.register(Context, Property, DataType, FileSystem)
def getone(  # noqa
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: FileSystem,
    *,
    id_: str,
):
    raise NotImplementedError


@commands.getone.register(Context, Property, File, FileSystem)
def getone(  # noqa
    context: Context,
    prop: Property,
    dtype: File,
    backend: FileSystem,
    *,
    id_: str,
):
    data = commands.getone(context, prop, prop.model.backend, id_=id_)
    if data is None:
        raise ItemDoesNotExist(prop, id=id_)
    data = (dtype.backend.path / data[prop.name]['_id']).read_bytes()
    return commands.cast_backend_to_python(context, prop, backend, data)



from starlette.requests import Request

from spinta import commands
from spinta.backends import get_select_prop_names
from spinta.backends import get_select_tree
from spinta.compat import urlparams_to_expr
from spinta.core.ufuncs import Expr
from spinta.renderer import render
from spinta.components import Context, Model, Property, Action, UrlParams
from spinta.backends import log_getall, log_getone
from spinta.types.datatype import DataType
from spinta.exceptions import NotFoundError, ItemDoesNotExist, UnavailableSubresource
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import flat_dicts_to_nested
from spinta.backends.postgresql.commands.query import PgQueryBuilder


@commands.getone.register(Context, Request, Model, PostgreSQL)
async def getone(
    context: Context,
    request: Request,
    model: Model,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)
    data = getone(context, model, backend, id_=params.pk)
    hidden_props = [prop.name for prop in model.properties.values() if prop.hidden]
    log_getone(context, data, select=params.select, hidden=hidden_props)
    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(context, model, action, select_tree)
    data = commands.prepare_data_for_response(
        context,
        Action.GETONE,
        model,
        backend,
        data,
        select=select_tree,
        prop_names=prop_names,
    )
    return render(context, request, model, params, data, action=action)


@commands.getone.register(Context, Model, PostgreSQL)
def getone(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    id_: str,
):
    connection = context.get('transaction').connection
    table = backend.get_table(model)
    try:
        result = backend.get(connection, table, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(model, id=id_)
    data = flat_dicts_to_nested(dict(result))
    data['_type'] = model.model_type()
    return commands.cast_backend_to_python(context, model, backend, data)


@commands.getone.register(Context, Request, Property, DataType, PostgreSQL)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: DataType,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    raise UnavailableSubresource(prop=prop.name, prop_type=prop.dtype.name)


@commands.getall.register(Context, Request, Model, PostgreSQL)
async def getall(
    context: Context,
    request: Request,
    model: Model,
    backend: PostgreSQL,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)
    expr = urlparams_to_expr(params)
    rows = getall(context, model, backend, query=expr)
    hidden_props = [prop.name for prop in model.properties.values() if prop.hidden]
    rows = log_getall(context, rows, select=params.select, hidden=hidden_props)
    if not params.count:
        select_tree = get_select_tree(context, action, params.select)
        prop_names = get_select_prop_names(context, model, action, select_tree)
        rows = (
            commands.prepare_data_for_response(
                context,
                action,
                model,
                backend,
                row,
                select=select_tree,
                prop_names=prop_names,
            )
            for row in rows
        )
    return render(context, request, model, params, rows, action=action)


@commands.getall.register(Context, Model, PostgreSQL)
def getall(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    query: Expr = None,
):
    assert isinstance(query, (Expr, type(None))), query
    connection = context.get('transaction').connection

    builder = PgQueryBuilder(context)
    builder.update(model=model)
    table = backend.get_table(model)
    env = builder.init(backend, table)
    expr = env.resolve(query)
    where = env.execute(expr)
    qry = env.build(where)

    conn = connection.execution_options(stream_results=True)
    result = conn.execute(qry)

    for row in result:
        row = flat_dicts_to_nested(dict(row))
        row = {
            '_type': model.model_type(),
            **row,
        }
        row = commands.cast_backend_to_python(context, model, backend, row)
        yield row
